import csv
import json
import logging
from pathlib import Path

import click
import pydantic

from contablo.csvimporter import import_csv_with_spec_detection
from contablo.csvtmplgen import CsvTemplateGenerator
from contablo.fields import FieldSpecRegistry
from contablo.fields import add_builtin_fieldspecs_to_registry
from contablo.importable import ImporTable
from contablo.importablemerge import LeftRightMatchRule
from contablo.importablemerge import importable_merge
from contablo.importspec import ImportSpec
from contablo.importspec import ImportSpecRegistry

logger = logging.getLogger(__file__)
log_levels = [logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG]


def fill_import_spec_registry(dir_or_file: str, registry: ImportSpecRegistry) -> None:
    import glob

    if not dir_or_file:
        return
    path = Path(dir_or_file)

    file_list = glob.glob((path / "*.json").as_posix()) if path.is_dir() else [dir_or_file]
    for import_config_file in file_list:

        try:
            with open(import_config_file) as f:
                data = json.load(f)
                if not all([key in data for key in ["label", "encoding", "type"]]):
                    logger.warning(f"Skipping {import_config_file}: Not a valid ImportSpec configuration.")
                    continue

                config = ImportSpec(**data)
                registry.add_import_spec(config, import_config_file)

        except json.JSONDecodeError:
            logger.warning(f"Skipping {import_config_file}: Not a valid JSON file.")
        except pydantic.ValidationError as e:
            print(f"Error when trying to initialize import spec from {import_config_file}:")
            from pprint import pprint

            pprint(e.errors())
        except Exception as e:
            logger.exception(e)


@click.group(context_settings=dict(help_option_names=["-h", "--help"]))
@click.option("-v", "--verbose", count=True)
def cli(verbose):
    """CLI tool to the ConTabLo python package."""
    logging.basicConfig(
        format="%(asctime)s.%(msecs)03d %(levelname)-8s [%(filename)s:%(lineno)d] %(message)s",
        datefmt="%Y-%m-%d_%H:%M:%S",
        level=log_levels[min(verbose, len(log_levels) - 1)],
    )


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.option("-v", "--verbose", count=True, default=None)
@click.option(
    "-t",
    "--target-spec",
    required=True,
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
    help="JSON file with target table specs.",
)
@click.option(
    "-c",
    "--config",
    type=click.Path(exists=True, file_okay=True, dir_okay=True),
    help="Location of the import spec. May be one single json file or a directory with several config files.",
)
@click.option(
    "-o",
    "--output-base",
    type=str,
    help="Write template to this directory (default: location of input file)",
)
@click.option(
    "-s/-S",
    "--samples/--no-samples",
    default=True,
    help="Decide wheter to include samples in the template or not",
)
@click.argument("csv-files", nargs=-1)
def mk_import_tmpl(
    verbose: int | None, csv_files: list[str], target_spec: str, config: str, output_base: str, samples: bool
):
    """Create an input configuration template for the given CSV file(s)."""
    if verbose is not None:
        logging.getLogger().setLevel(log_levels[min(verbose, len(log_levels) - 1)])

    fields = []

    if target_spec is not None:
        # Initialize target table's field specs from json file. See specs/fieldspec-banking.json for an example.
        # Refer to contablo.fields.ImportSpec subclasses for available types and attributes.
        with open(target_spec) as target_spec_file:
            fieldspecs = FieldSpecRegistry()
            add_builtin_fieldspecs_to_registry(fieldspecs)
            fields = fieldspecs.make_spec_list(json.load(target_spec_file))

    registry = ImportSpecRegistry()
    fill_import_spec_registry(config, registry)

    csv_files = list(csv_files)

    generator = CsvTemplateGenerator(fields, registry)
    generator.add_files(csv_files)
    generator.make_templates(output_path_base=output_base, skip_samples=not samples)


cli.add_command(mk_import_tmpl)


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.option("-v", "--verbose", count=True, default=None)
@click.option(
    "-o",
    "--output-base",
    type=str,
    help="Write schema to this directory (default: location of input file)",
)
@click.argument("schema", nargs=-1)
def mk_schema(verbose: int | None, output_base: str, schema: str):
    """Create a JSON schema for validation of JSON based config files."""
    # Todo: Add a way to also include valid field names in the schema
    if verbose is not None:
        logging.getLogger().setLevel(log_levels[min(verbose, len(log_levels) - 1)])

    schema: list[str] = list(set(schema))
    known_schemas: dict[str, pydantic.BaseModel] = {
        "ImportSpec": ImportSpec,
    }

    for sch in schema:
        if sch not in known_schemas:
            print(f"Unknown schema '{sch}', choose one of {known_schemas.keys()}")
            continue
        data = known_schemas[sch].model_json_schema()
        filename = f"{output_base}{sch.lower()}.json"
        print(f"Writing schema for <{known_schemas[sch].__name__}> to {filename} ...")
        with open(filename, "w") as f:
            json.dump(data, f, indent=2)


cli.add_command(mk_schema)


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.option("-v", "--verbose", count=True, default=None)
@click.option(
    "-t",
    "--target-spec",
    required=True,
    type=click.Path(exists=True, file_okay=True, dir_okay=False),
    help="JSON file with target table specs.",
)
@click.option(
    "-c",
    "--config",
    type=click.Path(exists=True, file_okay=True, dir_okay=True),
    help="Location of the import spec. May be one simngle json file or a directory with several config files.",
)
@click.option(
    "-o",
    "--output-file",
    type=str,
    help="Export merged imported data to this file.",
)
@click.argument("csv-files", nargs=-1, required=True, type=click.Path(exists=True, file_okay=True, dir_okay=False))
def convert(verbose: int | None, csv_files: list[str], target_spec: str, config: str, output_file: str):
    """Load the given CSV file(s) based on their configurations and write resulting table(s)."""
    if verbose is not None:
        logging.getLogger().setLevel(log_levels[min(verbose, len(log_levels) - 1)])

    registry = ImportSpecRegistry()
    fill_import_spec_registry(config, registry)

    # Initialize target table's field specs from json file. See specs/fieldspec-banking.json for an example.
    # Refer to contablo.fields.ImportSpec subclasses for available types and attributes.
    with open(target_spec) as target_spec_file:
        field_spec_registry = FieldSpecRegistry()
        add_builtin_fieldspecs_to_registry(field_spec_registry)
        fields = field_spec_registry.make_spec_list(json.load(target_spec_file))

    result = ImporTable(fields)
    for csv_file in csv_files:
        importable = import_csv_with_spec_detection(csv_file, registry, result.clone_empty, field_spec_registry)
        if not importable:
            print(f"--- importing from {csv_file} yields nothing ---")
            continue
        result = importable_merge(importable, result, [LeftRightMatchRule({}, ["imported_from"])])
        print(f"--- importing from {csv_file} with {len(importable)} entries results in {len(result)} after merge ---")

    if output_file is not None:
        print("Exporting merged data...")
        with open(output_file, "w") as csv_file:
            csv_writer = csv.writer(csv_file)
            csv_writer.writerows(result.get_flat_table(convert_func=str, fallback="", include_header=True))


cli.add_command(convert)



from __future__ import annotations

import logging

import magic
import pydantic

logger = logging.getLogger(__file__)


class UnknownEncodingError(Exception):
    pass


def get_file_encoding(filename: str) -> str:
    """tries to figure out the correct encoding of the given file"""

    with open(filename, "rb") as f:
        blob = f.read()
        m = magic.Magic(mime_encoding=True)
        encoding = m.from_buffer(blob)
        return "utf-8-sig" if encoding == "utf-8" else encoding

    raise UnknownEncodingError(f"Could not figure out encoding of '{filename}'.")


def load_chunked_textfile(
    filename: str,
    chunk_delimiters: list[str] = None,
    encoding: str = None,
) -> list[list[tuple[int, str]]]:
    """Read one or more chunks from the given file, delimited by empty lines or one of the specified delimiters."""
    chunk_delimiters = chunk_delimiters if chunk_delimiters is not None else ["", '""', "''"]
    chunks = [[]]
    encoding = encoding or get_file_encoding(filename)
    with open(filename, encoding=encoding) as f:
        chunk_num = 0
        for i, row in enumerate(f.readlines(), 1):
            if row.strip() in chunk_delimiters:
                logging.debug(f"New table in {filename} possibly starting at line {i}.")
                chunk_num += 1
                chunks.append([])
                continue
            chunks[chunk_num].append((i, row.strip()))

    logger.info(f"Found {len(chunks)} chunks in '{filename}'")
    return chunks


class ChunkInfo(pydantic.BaseModel):
    """Collects information on data chunks in files.

    A CSV file may contain a number of sections (chunks) delimited by an empty line.  Each chunk comprises information on
    the starting line in the file, the column delimiter, a number of column names and data lines.

    In some files, there may be some sort of meta-data stored as key-value-pair chunks preceding the main table.
    These kind of chunks shall receive a list of empty column names with the same length as the data columns.

    The datalines member may be added to when two ChunkINfo objects are merged.
    """

    delimiter: str
    first_line: int
    columns: list[str]
    datalines: list[str]

    def __eq__(self, other: ChunkInfo) -> bool:
        # chunks are comparable, if the lists column names are identical.
        if self.columns != other.columns:
            return False
        return True

    def add_datalines_from(self, other: ChunkInfo) -> None:
        if self != other:
            raise ValueError("Could not add lines from another format")
        self.datalines += other.datalines


class CsvFileInfo(pydantic.BaseModel):
    source_files: list[str]
    file_encoding: str
    chunk_info: list[ChunkInfo]

    def __eq__(self, other: CsvFileInfo) -> bool:
        if self.file_encoding != other.file_encoding:
            return False
        if self.chunk_info != other.chunk_info:
            return False
        return True

    def add_from(self, other: CsvFileInfo) -> None:
        if self != other:
            raise ValueError("Could not add lines from incompatible file")
        for i in range(0, len(self.chunk_info)):
            self.chunk_info[i].add_datalines_from(other.chunk_info[i])
        self.source_files += other.source_files



from __future__ import annotations

import logging
from typing import Any
from typing import Iterable

import pydantic

from contablo.csv_helper import CsvFileInfo

logger = logging.getLogger(__file__)


class StrictAttribBaseModel(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(extra="forbid")


class ImportMatchRule(StrictAttribBaseModel):
    rule: str
    formats: dict[str, str] = {}  # specify formats for matched fields that require one
    implies: dict[str, str] = {}  # values may append "/"+format where required.
    # Todo: would we need onlyif_all, onlyif_any, notiff_all, notif_any, and which would take precendence?
    onlyif: dict[str, str] = {}  # key: field, value: condition; all conditions must match
    # notif: dict[str, str] = {}  # key: field, value: condition


class ImportColumnSpec(StrictAttribBaseModel):
    """Defines how a certain column is to be used."""

    # Todo: sometimes, the field may depend on the value of other columns,
    #       so eventually we need some basic rule matching

    label: str  # the column label in the file, may be empty
    comment: str = ""
    field: str = ""  # may contain field1+field2, empty if unused  # Todo: Rename to field
    format: str = ""  # field specific configuration
    match: list[ImportMatchRule] | None = []
    map: dict[str, str] | None = {}
    ignore: list[str] | None = []
    samples: list[str] | None = []


class ImportSpec(StrictAttribBaseModel):
    """Specifies how a data table can be imported, i.e. transofrmed to generic import data.

    :param label:
    :param type: account|depot|mixed|snapshot
    :defaults: keys are field identifiers
    """

    label: str
    type: str
    encoding: str | None = None
    skip_lines: int = 0
    delimiter: str = ","
    defaults: dict[str, str] = {}  # values may append "/"+format where required.
    # Todo: formats: dict[str, str] = {}  # default formats for types (number, date, time, datetime) or specific column labels
    columns: list[ImportColumnSpec] = []
    fields: list[dict[str, str]] = []
    transforms: dict[str, str] = {}

    @property
    def column_labels(self):
        return [c.label for c in self.columns]

    def matches(self, other: CsvFileInfo) -> bool:
        """Test if this spec can by used to import the file described by the given csv file info."""
        logger.info("ImportSpec.matches")
        if isinstance(other, CsvFileInfo):
            logger.info(f"Checking {other.source_files}")
            if other.file_encoding != self.encoding:
                logger.info("Spec differs in encoding")
                return False
            for idx, chunk in enumerate(other.chunk_info, 1):
                if chunk.first_line != self.skip_lines + 1:
                    logger.info(f"Chunk {idx} differs in skip_lines")
                    continue
                if chunk.delimiter != self.delimiter:
                    logger.info(f"Chunk {idx} differs in delimiter")
                    continue
                if chunk.columns != self.column_labels:
                    logger.info(f"Chunk {idx} differs in column labels")
                    logger.debug(f"  {chunk.columns=}")
                    logger.debug(f"  {self.column_labels=}")
                    continue
                logger.info(f"Chunk {idx} matches")
                return True

        return False


class ImportSpecRegistry:
    def __init__(self) -> None:
        self.registry: list[ImportSpec] = []
        self.label_source_map: dict[str, str] = {}

    def iter_specs(self) -> Iterable[ImportSpec]:
        yield from self.registry

    def add_import_spec(self, spec: ImportSpec, source: str) -> None:
        self.registry.append(spec)
        self.label_source_map[spec.label] = source

    def query_by_file_info(self, info: CsvFileInfo) -> ImportSpec | None:
        logger.info(f"Scanning {len(self.registry)} specs for {info.source_files}")
        for spec in self.registry:
            logger.debug(f"Next is {spec.label}")
            if spec.matches(info):
                logger.debug(f"Found match: {spec.label}")
                return spec

    def query_source(self, label: str) -> str:
        return self.label_source_map.get(label, None)



from __future__ import annotations

import csv
import datetime
import logging

import dateparser

from contablo.codes import is_valid_isin
from contablo.numberformat import NumberFormat

logger = logging.getLogger(__file__)


def is_number(s: str) -> bool:
    """Test if the given string represents a python float or int number."""
    try:
        float(s)
        return True
    except ValueError:
        return False


def guess_separator(header: str) -> str:
    """Gues the most propable separator of a csv header"""
    s = csv.Sniffer().sniff(header)
    # logger.warning(f"{s}")
    return s.delimiter


# Todo: maybe get additional formats from some user configuration?

common_date_formats = {
    "yyyy-mm-dd": "%Y-%m-%d",
    "yyyy/mm/dd": "%Y/%m/%d",
    "yyyymmdd": "%Y%m%d",
    "dd.mm.yyyy": "%d.%m.%Y",
    "dd.mm.yy": "%d.%m.%y",
    "mm/dd/yyyy": "%m/%d/%Y",
    "mm/dd/yy": "%m/%d/%y",
    "dd/mm/yyyy": "%d/%m/%Y",
    "dd/mm/yy": "%d/%m/%y",
    "dd-mm-yyyy": "%d-%m-%Y",
    "dd-mm-yy": "%d-%m-%y",
    "mm-dd-yy": "%m/%d/%y",
    "mm-dd-yyyy": "%m/%d/%Y",
    "yyyy-mm": "%Y-%m",
    "yy-mm-dd": "%y-%m-%d",
    "dd mmm yyyy": "%d %B %Y",
}

common_time_formats = {
    "HH:MM:SS": "%H:%M:%S",
}

common_datetime_formats = {
    "dd.mm.yyyy HH:MM:SS": "%d.%m.%Y %H:%M:%S",
    "dd.mm.yyyy_HH:MM:SS": "%d.%m.%Y_%H:%M:%S",
    "yyyy-mm-dd HH:MM:SS": "%Y-%m-%d %H:%M:%S",
    "yyyy-mm-dd_HH:MM:SS": "%Y-%m-%d_%H:%M:%S",
    "dd mmm yyyy HH:MM:SS": "%d %B %Y %H:%M:%S",
}

DEFAULT_DATE = datetime.datetime.strptime("01:01:01", "%H:%M:%S").date()


def parse_datetime(text: str, datetime_format: str) -> datetime.datetime | None:
    """Parse text for a datetime conforming to the strptime compatible date_format.

    This will first try to use the builtin strptime which conforms more stritctly to the given format.
    Only if that fails, dateparser.parser is used to try parse strings that ptrptime couldn't handle, like
    localised date reprensentations in the form "01. Januar 2012" (german)
    """
    try:
        return datetime.datetime.strptime(text, datetime_format)
    except ValueError:
        pass

    # only try dateparser if there is actual text that hints towards month names:
    if not any(c.isalpha() for c in text) or "%B" not in datetime_format:
        return None
    try:
        return dateparser.parse(text, date_formats=[datetime_format])
    except ValueError as e:
        logger.debug(f"Failed to parse <{text}> with <{datetime_format}>: {e}")
        return None


def parse_date(text: str, date_format: str) -> datetime.date | None:
    """Parsed text for a date conforming to the strptime compatible date_format.

    Returns None instead of raising an expection
    """
    try:
        return parse_datetime(text, date_format).date()
    except (ValueError, AttributeError):
        return None


def parse_time(text: str, time_format: str) -> datetime.time | None:
    """Parsed text for a date conforming to the strptime compatible date_format.

    Returns None instead of raising an expection
    """
    try:
        return parse_datetime(text, time_format).time()
    except (ValueError, AttributeError):
        return None


def is_date_strptime(sample: str, strptime_format: str) -> bool:
    """Checks if the given date string conforms to the format using dd,MM,yy or yyy"""
    try:
        # avoid false success due to dateparser.parser being too lax:
        if ":" in sample:
            return None
        dt = parse_datetime(sample, strptime_format)
        return dt.date() != DEFAULT_DATE
    except (ValueError, AttributeError):
        pass

    return False


def is_time_strptime(sample: str, strptime_format: str) -> bool:
    """Checks if the given date string conforms to the format using dd,MM,yy or yyy"""
    try:
        # dt = datetime.datetime.strptime(sample, strptime_format)
        dt = parse_datetime(sample, strptime_format)
        return dt.date() == DEFAULT_DATE
    except (ValueError, AttributeError):
        pass

    return False


def is_datetime_strptime(sample: str, strptime_format: str):
    """Checks if the given date string conforms to the format using dd,MM,yy or yyy"""
    # avoid false positive due to dateparser.parser being too lax:
    if ":" not in sample:
        return False
    try:
        return parse_datetime(sample, strptime_format) is not None
    except ValueError:
        pass

    return False


def is_date(sample: str, date_format: str):
    """Checks if the given date string conforms to the format using dd,MM,yy or yyy"""
    if date_format not in common_date_formats:
        return False
    return is_date_strptime(sample, common_date_formats[date_format])


def is_time(sample: str, time_format: str):
    """Checks if the given date string conforms to the format using dd,MM,yy or yyy"""
    if time_format not in common_time_formats:
        return False
    return is_time_strptime(sample, common_time_formats[time_format])


def is_datetime(sample: str, datetime_format: str):
    """Checks if the given date string conforms to the format using dd,MM,yy or yyy"""
    if datetime_format not in common_datetime_formats:
        return False
    return is_datetime_strptime(sample, common_datetime_formats[datetime_format])


def get_date_strptime_from_format(date_format: str):
    assert (
        date_format in common_date_formats
    ), f"Unknown date format {date_format}, expecting one of {common_date_formats.keys()}"
    return common_date_formats.get(date_format)


def get_time_strptime_from_format(time_format: str):
    assert (
        time_format in common_time_formats
    ), f"Unknown time format {time_format}, expecting one of {common_time_formats.keys()}"
    return common_time_formats.get(time_format)


def get_datetime_strptime_from_format(datetime_format: str):
    assert (
        datetime_format in common_datetime_formats
    ), f"Unknown datetime format {datetime_format}, expecting one of {common_datetime_formats.keys()}"
    return common_datetime_formats.get(datetime_format)


def guess_date_format(samples: list[str]) -> str | None:
    for date_format, strptime_format in common_date_formats.items():
        if all([is_date_strptime(s, strptime_format) for s in samples]):
            return date_format
    return None


def guess_time_format(samples: list[str]) -> str | None:
    for time_format, strptime_format in common_time_formats.items():
        if all([is_time_strptime(s, strptime_format) for s in samples]):
            return time_format
    return None


def guess_datetime_format(samples: list[str]) -> str | None:
    for datetime_format, strptime_format in common_datetime_formats.items():
        if all([is_datetime_strptime(s, strptime_format) for s in samples]):
            return datetime_format
    return None


def guess_number_format(samples: list[str]) -> str | None:
    signs, thou_seps, frac_seps = set(), set(), set()
    integer_count = 0
    for sample in samples:
        try:
            format = NumberFormat.from_format(sample)
            if format.is_integer:
                integer_count += 1
                continue
            if format.thou_sep:
                thou_seps.add(format.thou_sep)
            if format.frac_sep:
                frac_seps.add(format.frac_sep)
            if format.sign:
                signs.add(format.sign)
        except ValueError:
            return None

    is_integer = integer_count == len(samples)

    if not is_integer:
        if len(thou_seps) > 1 or len(frac_seps) != 1:
            return None
        if len(thou_seps) > 1 and thou_seps == frac_seps:
            return None

    sign = "" if len(signs) == 0 else ("+" if "+" in signs else "-")
    if is_integer:
        format = NumberFormat(sign, "_", "")
    else:
        format = NumberFormat(sign, "" if len(thou_seps) == 0 else thou_seps.pop(), frac_seps.pop())

    if not all(map(format.is_valid_number, samples)):
        return None
    return format.format


def guess_field_and_format(samples: list[str]) -> tuple[str, str]:
    if not len(samples):
        return "empty", ""

    if all(map(is_valid_isin, samples)):
        return "isin", ""

    if (format := guess_date_format(samples)) is not None:
        return "date", format

    if (format := guess_time_format(samples)) is not None:
        return "time", format

    if (format := guess_datetime_format(samples)) is not None:
        return "datetime", format

    if (format := guess_number_format(samples)) is not None:
        return "number", format

    return "", ""


def format_implicit(tmpl: str, data: dict[str, str]) -> str:
    return tmpl.format_map(data)


def format_tmpl_str(tmpl: str, data: dict) -> str:
    if tmpl is None:
        return ""
    if not isinstance(data, dict):
        return tmpl
    for k, v in data.items():
        if isinstance(v, list):
            while f"{{{k}}}" in tmpl and len(v):
                val = v.pop(0)
                tmpl = tmpl.replace(f"{{{k}}}", val, 1)
            continue

        tmpl = tmpl.replace(f"{{{k}}}", f"{v}", 1)
    return tmpl



import iso3166


def is_luhn_valid(n: str) -> bool:
    # from https://stackoverflow.com/questions/21079439/implementation-of-luhn-formula
    r = [int(ch) for ch in str(n)][::-1]
    return (sum(r[0::2]) + sum(sum(divmod(d * 2, 10)) for d in r[1::2])) % 10 == 0


def is_valid_isin(isin: str) -> bool:
    # https://de.wikipedia.org/wiki/Internationale_Wertpapierkennnummer
    if len(isin) != 12:
        return False
    isin = isin.upper()
    cc = isin[0:2]
    if cc not in ["XS", "EU"]:
        try:
            iso3166.countries.get(cc)
        except KeyError:
            return False
    code = "".join([c if c.isdigit() else str(10 + ord(c) - ord("A")) for c in list(isin)])
    # print(isin, cc, nsin, cs, code)
    if not is_luhn_valid(code):
        return False
    return True



from __future__ import annotations

import logging
from typing import Any
from typing import Callable

import pydantic
from arithmetic_expressions import Expression

from contablo.fields import FieldSpec
from contablo.match import dicts_equal_in_keys

logger = logging.getLogger(__file__)


class ImportDatum(pydantic.BaseModel):
    source_lbl: str  # source column label
    raw_value: str
    format: str


class ImporTable:
    """Intermediary table for data to be imported.

    May contain additional metadata found in the import file.

    Import data of the same kind (aka using the same import spec) may be distributed over several
    files, possibly with duplicated entries. These duplicates must be removed in a first step,
    yielding one Importable instance with unique entries.

    Import files originating from the same exporting site may contain aspects of the same
    transactions, e.g. one file containing _value_ and the other _value+fees_, from which
    _fees_ can be calculated. Two or more such Importable instances may be merged yielding
    an Importable instance with unique combined entries.

    Each Importable instance with unique combined entries may be post-processed to interpret
    each entry as to how to create transactions for the main ledger.
    """

    def __init__(self, fields: list[FieldSpec]) -> None:
        self.fields_list: list[FieldSpec] = fields
        self.extra_fields_list: list[FieldSpec] = []
        self.transforms: dict[str, Expression] = {}
        self.data_vector: list[dict[str, Any]] = []  # see self.columns for valid keys

    @property
    def fields(self) -> dict[str, FieldSpec]:
        return {t.name: t for t in self.fields_list + self.extra_fields_list}

    @property
    def columns(self) -> list[str]:
        return [t.name for t in self.fields_list + self.extra_fields_list]

    @property
    def export_columns(self) -> list[str]:
        return [t.name for t in self.fields_list]

    def __len__(self) -> int:
        return len(self.data_vector)

    def clone_empty(self) -> ImporTable:
        return ImporTable(self.fields_list.copy())

    def add_extra_fields(self, fields: list[FieldSpec]) -> None:
        self.extra_fields_list = fields

    def add_transforms(self, transforms: dict[str, str]) -> None:
        self.transforms.update({k: Expression.parse(v) for k, v in transforms.items()})

    def get_columns(self) -> list[str]:
        return [c for c in self.columns]

    def get_data(self) -> list[dict[str, Any]]:
        # create a semi-deep copy:
        return [{k: v for k, v in row.items()} for row in self.data_vector]

    def get_flat_table(
        self,
        convert_func: Callable[[Any], str] | None = None,
        fallback: Any = None,
        include_header: bool = False,
    ) -> list[list[Any]]:
        rows = []
        if include_header:
            rows.append(self.columns)
        for entry in self.data_vector:
            row = []
            for column in self.columns:
                value = entry.get(column, None)
                if value is None:
                    value = fallback
                elif convert_func is not None:
                    value = convert_func(value)
                row.append(value)
            rows.append(row)

        return rows

    def iter_data(self, reversed: bool = False):
        yield from self.data_vector[:: -1 if reversed else 1]

    def compatible_fields(self, field_or_type: str) -> list[str]:
        field_map = {t.name: t for t in self.fields_list}
        if field_or_type in field_map:
            return [field_or_type]
        return [t.name for t in self.fields_list if t.type == field_or_type]

    def add(self, source: str, import_data: dict[str, ImportDatum]) -> None:
        """Add new dataset unless it contains a field "drop"."""
        # Todo: Error handling needs better design, e.g. specifiying source (e.g. "file:line")
        errors = []
        if import_data.get("drop", None) is not None:
            return
        for k, v in import_data.items():
            if k not in self.fields:
                errors.append(f"Unknown field <{k}>: {v}")
            if not isinstance(v, ImportDatum):
                errors.append(f"Implementation error: <{k}> requires type ImportDatum, got: {v}")
        if errors:
            print("** Errors:")
            for error in errors:
                print(f"   {error}")
            return
        data = {"imported_from": source}
        for field, datum in import_data.items():
            try:
                data[field] = self.fields[field].convert(datum.raw_value, datum.format)
            except (AssertionError, ValueError) as e:
                logger.exception(e)
                errors.append(f"{e} for {field=} and {datum=}")
        if errors:
            print("** Errors:")
            for error in errors:
                print(f"   {error}")
            raise ImportError("There were errors while adding data")  # raise a more appropriate Exception

        for column, expression in self.transforms.items():
            data[column] = self.evaluate(expression, data)

        self.data_vector.append(data)

    def evaluate(self, expression: Expression, data: dict[str, Any]) -> Any:
        """Evaluate the expression in a safe context, where missing values are replaced with zero."""
        context = {k: data.get(k, v.zero) for k, v in self.fields.items() if v.zero is not None}
        # Todo: package arithmetic_expressions is missing a method to start out with a clean context.
        # Since the arithmetic engine is a singleton and the default context is updated with each evaluate() call,
        # previous variables are remembered. This is a problem if a previous importable had a field defined which
        # the current importable is missing: The previous variable value is still stored in the context and used in
        # the expression without raising a arithmetic_expressions.engine.UndefinedVariableError
        expression.engine._default_context.clear()
        return expression.evaluate(**context)

    def merge_in(self, other: ImporTable):
        """Merge another importable into this one while dropping duplicates.

        For duplicate detection, see is_known_entry()"""
        for data in other.iter_data():
            if self.is_known_entry(data):
                continue
            self.data_vector.append(data)

    def is_known_entry(self, data: dict[str, Any]) -> bool:
        """Checks if the provided data is already known.

        Only keys in self.columns are checked; additional keys like imported_from are ignored.
        This allows to identify identical data imported from different files as duplicates.
        """
        # Todo: devise some caching
        for my_data in self.iter_data():
            if dicts_equal_in_keys(my_data, data, self.columns):
                return True
        return False

    # def to_dataframe(self) -> pd.DataFrame:  # on application level: DataFrame(imp.get_data(), columns=im.get_columns())
    #     columns = self.columns + ["imported_from"]
    #     df = pd.DataFrame(self.data_vector)
    #     for lbl in columns:
    #         if lbl not in df:
    #             df[lbl] = pd.NA
    #     return df[columns]



from __future__ import annotations

from dataclasses import dataclass


@dataclass
class NumberFormat:
    """Recognizes a number formatted with specific thousands and fractional separator."""

    # Todo: Decide on how to use sign.
    #   Currently, "-" allows for any sign, "+" enforces positive values
    #
    sign: str  #
    thou_sep: str  # thousands separator
    frac_sep: str  # fractional separator
    trailing_sign: bool = False

    @property
    def is_integer(self):
        return self.frac_sep == ""

    @property
    def format(self) -> str:
        elements = []
        if self.sign and not self.trailing_sign:
            elements.append(self.sign)
        elements.append("1")
        elements.append(self.thou_sep)
        elements.append("000")
        if self.frac_sep:
            elements.append(self.frac_sep)
            elements.append("00")
        if self.sign and self.trailing_sign:
            elements.append(self.sign)
        return "".join(elements)

    @staticmethod
    def from_format(format: str) -> NumberFormat:
        # Todo: This has become way too complicated, there should be a better way
        assert format != "" and format is not None, "Number requires a format"
        if not all([c in "+-.,'_0123456789" for c in format]):
            raise ValueError(f"Invalid character in '{format}'.")

        with_sign = format[0] in ["+", "-"]

        if with_sign:
            sign, number_fmt = format[0], format[1:]
        else:
            sign, number_fmt = "", format

        # fast track for integer values:
        try:
            int(number_fmt, 10)
            return NumberFormat(sign, "_", "")
        except ValueError:
            pass

        seps = [c for c in number_fmt if not c.isdigit()]

        if len(seps) == 1:
            thou_sep, frac_sep = "", seps[0]
        elif len(seps) == 2:
            thou_sep, frac_sep = seps
            if thou_sep == frac_sep:
                raise ValueError(
                    f"A valid number format requires maximal one fractional delimiter, not like '{format}'"
                )
        else:
            # there can be multiple thousands separators but only one in the format
            sepchars = list(set(seps))
            if len(sepchars) != 2:
                raise ValueError(f"More than two different separators in '{format}'.")
            if seps.count(sepchars[0]) == 1:
                thou_sep, frac_sep = sepchars[1], sepchars[0]
            elif seps.count(sepchars[1]) == 1:
                thou_sep, frac_sep = sepchars[0], sepchars[1]
            else:
                raise ValueError(f"There's more than a single decimal separator in '{format}'.")

        return NumberFormat(sign, thou_sep, frac_sep)

    def is_valid_number(self, sample: str, /, raise_on_fail: bool = False):
        """Check if the given number conforms to the format"""
        # if not all([c in "+-.,'_0123456789" for c in sample]):
        #     return False
        try:
            sample_format = NumberFormat.from_format(sample)
            if (sample_format.frac_sep, sample_format.thou_sep) == (self.thou_sep, ""):
                sample_format.frac_sep, sample_format.thou_sep = self.frac_sep, self.thou_sep

        except ValueError as e:
            if raise_on_fail:
                raise ValueError(f"Fractional separator not matched in {sample}: {e}")
            return False

        if sample_format.is_integer:
            return True

        # we expect the same fractional seperator but accept an integer number
        if sample_format.frac_sep and (sample_format.frac_sep != self.frac_sep):
            if raise_on_fail:
                raise ValueError(f"Fractional separator mismatch: {self.frac_sep} != {sample_format.frac_sep}")
            return False

        # numbers smaller than 1000 will not contain a thousands separator
        if sample_format.thou_sep and (sample_format.thou_sep != self.thou_sep):
            if raise_on_fail:
                raise ValueError(f"Thousands separator mismatch: {self.thou_sep} != {sample_format.thou_sep}")
            return False

        # Relax a little on the sign thingy.
        # We want to keep tracvk of signs but won't enforce identical signs
        #
        # if sample_format.sign and not self.sign:
        #     return False

        # last check: there need to be exactly 3 digits betwenn two separators
        if self.thou_sep:
            if self.frac_sep and self.frac_sep in sample:
                full, frac = sample.replace("+", "").replace("-", "").split(self.frac_sep, 1)
            else:
                full, frac = sample.replace("+", "").replace("-", ""), ""
            full_parts = full.split(self.thou_sep)
            frac_parts = frac.split(self.thou_sep)

            if len(full_parts) > 1 and not all([len(p) == 3 for p in full_parts[1:]]):
                if raise_on_fail:
                    raise ValueError(f"Wrong number of digits left of {self.frac_sep}: {full_parts[1:]}")
                return False
            if len(frac_parts) > 1 and not all([len(p) == 3 for p in frac_parts[0:-1]]):
                if raise_on_fail:
                    raise ValueError(f"Wrong number of digits right of {self.frac_sep}: {frac_parts[0:-1]}")
                return False

        return True

    def normalize(self, number: str) -> str:
        """Converts the given number to a format that can be casted to float or decimal."""
        if not self.is_valid_number(number):
            raise ValueError(f"Number {number} does not conform to format {self.format}.")

        is_negative = "-" in number
        number = number.replace("+", "").replace("-", "").strip()
        if is_negative:
            number = "-" + number
        if self.thou_sep:
            number = number.replace(self.thou_sep, "")
        if self.frac_sep:
            number = number.replace(self.frac_sep, ".")

        return number



import logging
from dataclasses import dataclass
from decimal import Decimal
from typing import Any
from typing import Protocol

from contablo.importable import ImporTable

logger = logging.getLogger(__file__)


class AssetLookup(Protocol):
    def lookup_symbol(isin_or_wkn: str) -> str: ...


def is_undef(value: Any, ignore_undef):
    if not ignore_undef:
        return False
    for item in ignore_undef:
        if isinstance(value, Decimal):
            return Decimal is None
        if isinstance(value, type(item)):
            if value == item:
                return True
    return False


def dicts_match_by_map(
    left: dict[str, Any],
    right: dict[str, Any],
    match_map: dict[str, str] = None,
    ignore_left: list[str] = None,
    ignore_right: list[str] = None,
    ignore_undef: list[Any] = None,
) -> bool:
    """Check if two dicts match, considering a subset of dict keys.
    Keys can be mapped between the two dicts.
    Remaining keys will be checked for same values in both dicts unless mentioned in ignore lists.
    Optionally, individual (non-mapped) keys may be missing on either side if value matches ignore_undef
    """
    if match_map is None:
        raise ValueError("match_map cannot be empty.")
    ignore_left = ignore_left or []
    ignore_right = ignore_right or []
    if any([k not in right for k in match_map.keys()]):
        # logger.debug(f"missing at least one of {match_map.keys()} in {right.keys()}")
        return False
    if any([v not in left for v in match_map.values()]):
        # logger.debug(f"missing at least one of {match_map.values()} in {left.keys()}")
        return False
    for right_key, left_key in match_map.items():
        if left.get(left_key, None) != right.get(right_key, None):
            # logger.debug(
            #     f"Mismatch 1: left[{left_key}]={left.get(left_key, None)} and right[{right_key}]={right.get(right_key, None)}"
            # )
            return False

    for key in left.keys():
        if key in ignore_left:
            continue
        if key in match_map.keys() or key in match_map.values():
            continue
        if ignore_undef and key not in right:
            continue
        if is_undef(left[key], ignore_undef):
            continue
        if key in right and is_undef(right[key], ignore_undef):
            continue
        if left.get(key, None) != right.get(key, None):
            logger.debug(f"Mismatch 2: left[{key}]={left.get(key, None)} and right[{key}]={right.get(key, None)}")
            return False
    for key in right.keys():
        if key in ignore_right:
            continue
        if key in match_map.keys() or key in match_map.values():
            continue
        if ignore_undef and key not in left:
            continue
        if is_undef(right[key], ignore_undef):
            continue
        if key in left and is_undef(left[key], ignore_undef):
            continue
        if left.get(key, None) != right.get(key, None):
            logger.debug(f"Mismatch 3: left[{key}]={left.get(key, None)} and right[{key}]={right.get(key, None)}")
            return False

    return True


def pick_one(
    left: list[dict[str, Any]],
    right: dict[str, Any],
    match_map: dict[str, str],
    ignore_left: list[str] = None,
    ignore_right: list[str] = None,
    ignore_values: list[Any] = None,
    *,
    remove_match: bool = False,
) -> dict[str, Any]:
    """Try to find an entry in the left list matching the right dict.

    match_map dictates, which keys from "right" (key) must match keys in "left" (value).
    """
    match_idx = set()
    for idx, row in enumerate(left):
        if dicts_match_by_map(row, right, match_map, ignore_left, ignore_right, ignore_values):
            match_idx.add(idx)
    # print(f"{match_idx=}")
    if len(match_idx) == 0:
        # logger.warning(f"No match for {right}, ignoring {ignore_right}.")
        return {}
    if len(match_idx) > 1:
        logger.warning(f"multiple matches: {match_idx=}")
        return {}
    logger.debug(f"Found match at idx {idx}")
    idx = match_idx.pop()
    result = left[idx].copy()
    if remove_match:
        del left[idx]
    return result


@dataclass
class LeftRightMatchRule:
    left_right_map: dict[str, str]
    ignored_fields: list[str]


def already_in(from_data: dict[str, str], to_data: dict[str, str], ignore_keys: list[str], undef: list[str]):
    for key, value in from_data.items():
        if key in ignore_keys:
            continue
        if is_undef(value, undef):
            continue
        if key not in to_data:
            # print(f"  not yet in: {key}")
            return False
        if is_undef(to_data[key], undef):
            continue
        if to_data[key] != value:
            # print(f"  not yet in: {key}={value}")
            return False
    return True


def importable_merge_two(
    source: ImporTable,
    target: ImporTable,
    match_rules: list[LeftRightMatchRule],
    addable_fields: list[str],
) -> ImporTable:
    """Merge two importables, using a set of match rules to identify mergable entries. Not suitable for inner merges."""
    # this implementation consumes all merable items from both importables to produce new entries.
    # finally, all non-mergable items from both importables are also added.
    imp = target.clone_empty()

    tgt = target.data_vector.copy()
    src = source.data_vector.copy()
    undef = [None, ""]

    def try_merge(match_map: dict[str, str], ignore_keys: list[str]) -> None:
        rem = []
        if not tgt:
            # print(f"Already finished, skipping {match_map}")
            return src
        for row in src:
            ignored_keys = [k for k in ignore_keys]
            # logger.warning(f"#### {row=}")
            if row.get("_allow_add", False):
                ignored_keys.extend(addable_fields)
                # print(f"***** {ignored_keys}")
            match = pick_one(tgt, row, match_map, ignored_keys, ignored_keys, ignore_values=undef, remove_match=True)
            if not match:
                rem.append(row)
                continue
            if not already_in(row, match, ignored_keys, undef):
                msrc = "|".join([s for s in [match.get("imported_from", None), row.get("imported_from", None)] if s])
                # print(msrc)
                match.update(
                    {k: v for k, v in row.items() if not is_undef(v, undef) and is_undef(match.get(k, None), undef)}
                )
                match["imported_from"] = msrc
            imp.data_vector.append(match)
        return rem

    for match_rule in match_rules:
        if not tgt:
            break

        src = try_merge(match_rule.left_right_map, match_rule.ignored_fields)

    # Todo: find better means of debugging without adding pandas dependency
    # print()
    # if imp.data_vector:
    #     print("result:")
    #     print("\n".join(pd.DataFrame(imp.data_vector)["imported_from"].to_list()))
    # else:
    #     print("strange, nothing was merged...")
    # print()
    # if tgt:
    #     print("remaining target:")
    #     print("\n".join(pd.DataFrame(tgt)["imported_from"].to_list()))
    # else:
    #     print("all of target was comsumed")
    # print()
    # if src:
    #     print("remaining source:")
    #     print("\n".join(pd.DataFrame(src)["imported_from"].to_list()))
    # else:
    #     print("all of source was comsumed")

    for row in tgt:
        imp.data_vector.append(row)

    for row in src:
        imp.data_vector.append(row)

    return imp


def importable_merge_one(
    target: ImporTable,
    item: dict[str, Any],
    match_rules: list[LeftRightMatchRule],
    addable_fields: list[str],
) -> ImporTable:
    """Merge two importables, using a set of match rules to identify mergable entries. Not suitable for inner merges."""
    # this implementation consumes all merable items from the target importable to produce a new entry.
    # finally, all non-mergable items from both importables are also added.
    imp = target.clone_empty()

    tgt = target.data_vector.copy()
    undef = [None, ""]

    for match_rule in match_rules:
        match_map, ignore_keys = match_rule.left_right_map, match_rule.ignored_fields
        if not tgt:
            break
        ignored_keys = [k for k in ignore_keys]  # Todo: what???
        # logger.warning(f"#### {item=}")
        if item.get("_allow_add", False):
            ignored_keys.extend(addable_fields)
            # print(f"***** {ignored_keys}")
        match = pick_one(tgt, item, match_map, ignored_keys, ignored_keys, ignore_values=undef, remove_match=True)
        if not match:
            continue
        if not already_in(item, match, ignored_keys, undef):
            msrc = "|".join([s for s in [match.get("imported_from", None), item.get("imported_from", None)] if s])
            # print(msrc)
            match.update(
                {k: v for k, v in item.items() if not is_undef(v, undef) and is_undef(match.get(k, None), undef)}
            )
            match["imported_from"] = msrc
        elif item.get("_allow_add", False):
            for key in addable_fields:
                if match.get(key, None) is not None and item.get(key, None) is not None:
                    match[key] = match[key] + item[key]
        tgt.append(match)
        item = None
        break

    # if there was no match, ust append the item as is
    if item is not None:
        tgt.append(item)

    imp.data_vector = tgt

    return imp


def importable_merge(
    source: ImporTable,
    target: ImporTable,
    match_rules: list[LeftRightMatchRule] = None,
    addable_fields: list[str] = None,
) -> ImporTable:
    result = target
    for item in source.iter_data():
        result = importable_merge_one(result, item, match_rules or [], addable_fields or [])
    return result



import logging

from contablo.csv_helper import load_chunked_textfile
from contablo.fields import FieldSpecRegistry
from contablo.format_helpers import format_implicit
from contablo.format_helpers import guess_separator
from contablo.importable import ImporTable
from contablo.importable import ImportDatum
from contablo.importspec import ImportColumnSpec
from contablo.importspec import ImportSpec
from contablo.importspec import ImportSpecRegistry
from contablo.match import check_conditions
from contablo.match import match_to_template

logger = logging.getLogger(__file__)


class ImportSpecError(Exception):
    pass


class ImportColumnMismatchError(ImportSpecError):
    pass


class ImportSpecExceededError(ImportSpecError):
    """Raised when an input value is not covered by the given import spec"""

    pass


def make_field_import_datum_dict(
    src_lbl: str,
    spec: dict[str, str],
    fmt_sep: str,
    row_dict: dict[str, str] | None = None,
) -> dict[str, ImportDatum]:
    result = {}
    for field, spec in spec.items():
        value, format = spec.split(fmt_sep, 1) if fmt_sep in spec else (spec, "")
        if row_dict is not None:
            value = format_implicit(value, row_dict)
        result[field] = ImportDatum(source_lbl=src_lbl, raw_value=value, format=format)
    return result


def import_csv_with_spec_detection(
    csv_file: str,
    import_spec_registry: ImportSpecRegistry,
    importable_factory: ImporTable,
    field_spec_registry: FieldSpecRegistry,
) -> ImporTable | None:
    result: ImporTable = None
    found_specs = set()
    for spec in import_spec_registry.iter_specs():
        found = None
        try:
            found = import_csv_with_spec(csv_file, spec, importable_factory, field_spec_registry)
        except ImportColumnMismatchError:
            pass
        except Exception as e:
            logger.exception(e)
            print(f"Exception trying {spec.label} on {csv_file}: {e}")
        if not found:
            continue
        result = found
        found_specs.add(spec.label)

    if len(found_specs) == 0:
        print(f"Found no match for {csv_file}")
        return None

    elif len(found_specs) > 1:
        print(f"** Error: More than one specs matches {csv_file}.")
        print(f"** Error: Please restrict to one of {', '.join(found_specs)}")
        return None

    return result


def import_csv_with_spec(
    csv_file: str,
    import_spec: ImportSpec,
    importable_factory: ImporTable,
    registry: FieldSpecRegistry,
) -> ImporTable | None:
    """Import a single csv file with the given spec

    If the file content does not match the spec, the import will fail
    and another specs might be required to succeed.
    """
    chunks = load_chunked_textfile(csv_file)
    for i, chunk in enumerate(chunks, 1):  # chunk is a list of tuples comprising line number and content
        if not len(chunk):
            logging.debug(f"Chunk #{i:2d} is empty.")
            continue

        lines = [line for _, line in chunk]
        first = lines[0] if len(lines[0]) < 120 else f"{lines[0][:57]} [..] {lines[0][-57:]}"
        logging.debug(f"Chunk #{i:2d} comprises {len(lines)} lines; 1st line is:")
        logging.debug(f"          {first}")
        try:
            import csv

            reader = csv.reader(lines, delimiter=guess_separator(first), quoting=1)
            columns = next(reader)
            if columns != [cspec.label for cspec in import_spec.columns]:
                raise ImportColumnMismatchError()
            logging.info("Columns match, proceeding with import.")

            importable: ImporTable = importable_factory()
            importable.add_extra_fields(registry.make_spec_list(import_spec.fields))
            importable.add_transforms(import_spec.transforms)

            logging.debug(f"{columns=}")

            # Todo: Figure out a way to keep track of errors and warnings, including invalid lines
            for line, row in enumerate(reader, 2):
                add_to_importable_using_import_spec(importable, import_spec, row, f"{csv_file.split('/')[-1]}:{line}")

            return importable

        except NotImplementedError:  # Todo: leftover from previous implementation - keep or drop?
            raise

    return None


def add_to_importable_using_import_spec(
    importable: ImporTable,
    import_spec: ImportSpec,
    row: list[str],
    source: str,
) -> None:
    """Add data to an importable object from a single row in a source described by the given import_spec."""
    # Todo: Figure out a way to keep track of errors and warnings, including invalid lines - maybe return some log object?
    logging.debug(f"{source}: {row}")

    columns = [col.label for col in import_spec.columns]
    row_dict = {k: v for k, v in zip(columns, row)}
    columns_resolved: set[str] = set()  # needs to keep track of all columns with valid data source
    columns_mapped: set[str] = set()

    #
    # step 1: use import spec to collect fields and formats; do not yet handle match clauses
    #
    # do not fill in defaults, yet - otherwise we get a warning when a default value is overwritten
    field_data: dict[str, ImportDatum] = {}
    raw: str
    import_spec: ImportColumnSpec
    ignore_labels = []
    for raw, spec in zip(row, import_spec.columns):
        if not raw:
            ignore_labels.append(spec.label)
            columns_resolved.add(spec.label)
            continue
        if spec.ignore and raw in spec.ignore:  # checked before empty to detect unknown ignorable values
            ignore_labels.append(spec.label)
            columns_resolved.add(spec.label)
            continue
        label, format, field = spec.label, spec.format, spec.field
        if not field:
            # column may still provide data through map or match rules
            continue
        if field == "empty":
            if raw:
                print(f"** Error: Expecting <{label}> to be empty, but got <{raw}>")
            ignore_labels.append(spec.label)
            columns_resolved.add(spec.label)
            continue
        if field in field_data:
            prev = field_data[field].source_lbl
            print(f"** Warning: column {label} redefines field {field} already defined by {prev}")
        if spec.map:
            if raw in spec.map:
                from_raw, raw = raw, spec.map.get(raw)
                logger.debug(f"mapping {source}/{spec.label} from {from_raw} to {raw}")
                columns_mapped.add(spec.label)

        field_data[field] = ImportDatum(source_lbl=label, raw_value=raw, format=format)
        columns_resolved.add(label)

    logging.debug(f"***** {field_data=}")

    #
    # step 2: handle match clauses separately. only-if statements may only use data from step 2
    #
    match_results: dict[str, ImportDatum] = {}
    for raw, spec in zip(row, import_spec.columns):
        if spec.label in ignore_labels:
            continue
        match_groups: list[dict[str, str]] = []
        if not spec.match:
            continue
        for rule_idx, rule in enumerate(spec.match):
            data = match_to_template(raw, rule.rule)
            if data is None:  # None is no match, {} is a match but without data (e.g. with implies)
                logger.warning(f"no data for {raw=} {rule=}")
                continue  # this is normal, only one rule should match
            logging.debug(f"    ! found match with {data=}")
            if rule.onlyif:
                field_dict = {k: v.raw_value for k, v in field_data.items()}
                if not check_conditions(rule.onlyif, field_dict, row_dict):
                    print(f"** Warning: dropping match #{rule_idx} due to onlyif condition not met.")
                    continue
            match_data = {}
            for k, v in data.items():
                match_data[k] = ImportDatum(source_lbl=k, raw_value=v, format=rule.formats.get(k, ""))
            for k, v in make_field_import_datum_dict("(matched rule)", rule.implies, "/", row_dict).items():
                # v.raw_value = format_implicit(v.raw_value, row_dict)  # Todo: if still required, find a better way
                match_data[k] = v
            match_groups.append(match_data)
            # logging.debug(f"      -> {match_groups}")

        if (
            spec.match and not match_groups and spec.label not in columns_mapped
        ):  # and spec.label not in columns_resolved:
            print(f"** {spec.match=}")
            print(f"** {match_groups=}")
            raise ImportSpecExceededError(f"Input spec for {source}/{spec.label} does not cover value; {raw}")

        #
        # step 3: check for clashes between all match clauses
        #
        if len(match_groups):
            if len(match_groups) > 1:
                print(f"** Warning: Multiple matches in {source} column {spec.label}. Will use first match.")
            # logging.debug("      -> {match_groups[0]}")
            for field, value in match_groups[0].items():
                if field in match_results:
                    prev = match_results[field].source_lbl
                    print(f"** Warning: matched rule redefines field {field} already defined by {prev}")
                match_results[field] = value

    #
    # step 4: merge data from step 3 into step 4, starting with defaults
    #

    merged_data = make_field_import_datum_dict("(defaults)", import_spec.defaults, "/", row_dict)
    merged_data.update(field_data)
    merged_data.update(match_results)
    # logging.debug("    >", merged_data)

    #
    # step 5: pass field/value/format tuples into importable
    # importable.feed_data(series.to_dict(), spec)
    #
    importable.add(f"{import_spec.label}:{source}", merged_data)


class ImportColumnConfigError(Exception):
    pass



import csv
import datetime
import json
import logging
import time
from pathlib import Path
from typing import Protocol

from contablo.csv_helper import ChunkInfo
from contablo.csv_helper import CsvFileInfo
from contablo.csv_helper import get_file_encoding
from contablo.csv_helper import load_chunked_textfile
from contablo.fields import FieldSpec
from contablo.format_helpers import guess_field_and_format
from contablo.format_helpers import guess_separator
from contablo.importable import ImporTable
from contablo.importspec import ImportColumnSpec
from contablo.importspec import ImportSpec

logger = logging.getLogger(__file__)


class KnowsImportSpec(Protocol):
    def query_by_file_info(self, info: CsvFileInfo) -> ImportSpec: ...


class CsvTemplateGenerator:

    def __init__(self, fields: list[FieldSpec], import_specs_registry: KnowsImportSpec = None) -> None:
        # Todo: accept a repository with known import specs - ???
        self.fields = fields
        self.import_specs_registry = import_specs_registry
        self.input_formats: list[CsvFileInfo] = []
        self.input_specs: dict[str, list[str]] = {}  # maps exisiting ImportSpecs.labels to filenames

    def add_file_info(self, info: CsvFileInfo, ignore_known_specs: bool = False) -> None:
        """Add a CvsFileInfo to the internal list, or merge its filenames and datalines into exisiting ones.

        Existing specs known through the import_specs_registry are ignored and the filenames are instead noted for reporting.
        This behavior can be suppressed by asserting ignore_known_specs.
        """
        assert len(info.source_files) == 1
        if not ignore_known_specs and self.import_specs_registry is not None:
            logger.debug(f"Checking for known specs matching {info.source_files}")
            if spec := self.import_specs_registry.query_by_file_info(info):
                self.input_specs[spec.label] = self.input_specs.get(spec.label, []) + info.source_files
                logger.info(f"Format is already known as import spec <{spec.label}>")
                return

        for format in self.input_formats:
            logger.debug(f"compare file info to that from {format.source_files[0]}")
            if info == format:
                logger.info("Adding more data to previous format")
                format.add_from(info)
                return

        logger.info("Adding new format")
        self.input_formats.append(info)

    def add_files(self, csv_files: list[str]) -> None:
        for csv_file in csv_files:
            self.add_file(csv_file)

    def add_file(self, csv_file: str) -> None:
        """Add a csv file containing transactions to be analyzed.
        Files will be grouped by their csv properties including column labels."""
        encoding = get_file_encoding(csv_file)
        chunk_info = []
        chunks = load_chunked_textfile(csv_file, encoding=encoding)
        # Step 1: analyse chunks, extract separator and column labels, remember line index of header

        for chunk_idx, chunk in enumerate(chunks, 1):
            if not len(chunk):
                print(f"Chunk #{chunk_idx:2d} is empty.")
                continue

            logger.info(f"Chunk #{chunk_idx:2d} starts at line {chunk[0][0]} of {encoding}-encoded file {csv_file}")

            lines = [line for _, line in chunk]

            delimiter = guess_separator(lines[0])

            reader = csv.reader(lines, delimiter=delimiter, quoting=1)
            columns = next(reader)

            # detecting meta-data like this breaks easily: most of te time, there are only two fields (or 3 with the last being empty)
            # but at times, there can be entries in a metadata chunk with more fields
            # maybe it is more feasable to just assume that anything but the last chunk contains metadata rather than a real table.
            if len(columns) == 2 or (len(columns) == 3 and columns[-1] in ["", "''", '""']):
                # this must be metadata
                chunk_info.append(
                    ChunkInfo(delimiter=delimiter, first_line=chunk[0][0], columns=[""] * len(columns), datalines=lines)
                )
            else:
                chunk_info.append(
                    ChunkInfo(delimiter=delimiter, first_line=chunk[0][0], columns=columns, datalines=lines[1:])
                )

        self.add_file_info(CsvFileInfo(source_files=[csv_file], file_encoding=encoding, chunk_info=chunk_info))

    def make_templates(self, output_path_base: str = None, skip_samples: bool = False) -> None:
        """Output import spec template files based on the alanyzed csv files."""
        print(f"Found {len(self.input_formats)} distict file formats:")
        for idx, format in enumerate(self.input_formats, 1):
            print(
                f"  #{idx:2d}: {len(format.source_files)} files, {format.file_encoding} encoded, with {len(format.chunk_info)} chunks."
            )
            for ch, chunk in enumerate(format.chunk_info, 1):
                columns, lines = chunk.columns, chunk.datalines
                print(" " * 7 + f"#{ch:2d}: {len(columns)} columns, {len(lines)} samples")
                if not lines:
                    continue
                columns = chunk.delimiter.join(columns)
                head = columns if len(columns) < 120 else f"{columns[:57]} [..] {columns[-57:]}"
                first = lines[0] if len(lines[0]) < 120 else f"{lines[0][:57]} [..] {lines[0][-57:]}"
                print(" " * 12 + head)
                print(" " * 12 + first)

            spec = self.make_import_spec(format, self.fields, skip_samples=skip_samples)
            if output_path_base is None:
                print("  Not writing output file.")
                continue
            ts = datetime.datetime.fromtimestamp(time.time()).strftime("%Y%m%d_%H%M%S")
            path = Path(output_path_base)
            if path.is_dir():
                output_file = (path / f"new_import_spec-{ts}-{idx:02d}.json").as_posix()
            else:
                output_file = f"{output_path_base}-{ts}-{idx:02d}.json"
            try:
                with open(output_file, "w", encoding="utf8") as f:
                    print("// Generated with CsvTemplateGenerator from these files:", file=f)
                    for filename in format.source_files:
                        print(f"// - {filename}", file=f)
                    # avoid writing \u escaped chars
                    # see https://stackoverflow.com/a/18337754
                    json.dump(spec.model_dump(exclude_unset=True), f, indent=2, ensure_ascii=False)
                print(f"  Template was written to {output_file}")
            except Exception as e:
                print(f"  Failed to write to {output_file}: {e}")

    def make_import_spec(
        self,
        file_info: CsvFileInfo,
        fields: list[FieldSpec],
        skip_samples: bool = False,
    ) -> ImportSpec | None:
        if not file_info.chunk_info:
            return None
        chunk = file_info.chunk_info[-1]

        try:
            # collect samples for each column:
            reader = csv.reader(chunk.datalines, delimiter=chunk.delimiter, quoting=1)
            column_values = {k: set() for k in chunk.columns}
            for row in reader:
                for label, value in zip(chunk.columns, row):
                    if value == "":
                        continue
                    column_values[label].add(value)

            # try to figure out a SpecField type and possible field labels for each column:
            specs = []
            for label in chunk.columns:
                samples = list(column_values[label])
                field_type, format = guess_field_and_format(samples)
                spec = dict(label=label, field=field_type)
                field_hint = ImporTable(fields).compatible_fields(field_type or "string")
                if field_hint:
                    spec["field"] = "|".join(field_hint)

                spec["format"] = format
                if not field_type and not format and not skip_samples:
                    spec["samples"] = list(sorted(samples))

                specs.append(ImportColumnSpec(**spec))

            spec = ImportSpec(  # Todo: Use dependency injection
                label="INSTITUTION-SUBACCOUNT",
                type="account|depot|mixed|snapshot",  # Todo: Application specific, should be used with ImportSpec subclass
                encoding=file_info.file_encoding,
                skip_lines=chunk.first_line - 1,
                defaults={
                    "quote_currency": "EUR"
                },  # Todo: Application specific, should be used with ImportSpec subclass
                columns=specs,
            )
            return spec

        except NotImplementedError:  # Todo: leftover from previous implementation - keep or drop?
            pass



from __future__ import annotations

import datetime
import logging
from dataclasses import dataclass
from dataclasses import field
from decimal import Decimal
from typing import Any
from typing import Protocol

from contablo.format_helpers import common_date_formats
from contablo.format_helpers import common_datetime_formats
from contablo.format_helpers import common_time_formats
from contablo.format_helpers import is_number
from contablo.format_helpers import parse_datetime
from contablo.numberformat import NumberFormat

logger = logging.getLogger(__file__)


class FieldSpec(Protocol):
    """Interface for a field type specification.

    A field type specification is a class representing a field type that can be converted from a string.
    Objects of a class that implement this protocol have a property name (aka table column name) and a type
    description (e.g. "string" for the StringFieldSpec or "number" for the DecimalFieldSpec).

    The class method convert() allows to convert a string to the native type (e.g. str for StringFieldSpec or
    Decimal for StringFieldSpec), according to an optional format string that informs about how to interpret
    the textual value; see DateFieldSpec or DecimalFieldSpec for examples.  Types that are unsuitable for numeric
    expressions shall set the zero poperty to None.

    Types that are applicable to be used in expressions shall have a "zero" property of the respective type
    representing a zero value, e.g. int(0) or Decimal("0.0").

    Additinal properties may be required for each DieldSpec class to properly initialize and use a class object.
    """

    name: str
    type: str
    zero: Any

    @staticmethod
    def convert(value: str, format: str) -> Any:
        """Convert the text value from provided string format (if applicable) to native type.

        Return type may be None if conversion fails.
        """
        pass


def assert_field_spec_class(cls) -> bool:
    assert isinstance(cls, type), f"{cls} needs to be a class, not an object"
    assert isinstance(getattr(cls, "type", None), str), f"{cls} requires a member named 'type' of type 'str'"
    assert getattr(cls, "convert", None).__class__.__name__ == "function", f"{cls} requires a function 'convert'."


def is_field_spec_class(cls) -> bool:
    try:
        assert_field_spec_class(cls)
        return True
    except AssertionError:
        pass

    return False


@dataclass
class StringFieldSpec:
    name: str
    help: str
    type: str = field(default="string", init=False)
    zero: None = field(default=None, init=False)

    @staticmethod
    def convert(value: str, format: str) -> str:
        return str(value)


@dataclass
class EnumFieldSpec:
    name: str
    help: str
    items: list[str]
    type: str = field(default="enum", init=False)
    zero: None = field(default=None, init=False)

    def convert(self, value: str, format: str) -> str:
        assert value in self.items, f"Unknown item <{value}>, expecting of of {self.items}"
        return value


@dataclass
class IntFieldSpec:
    name: str
    help: str
    type: str = field(default="integer", init=False)
    zero: int = field(default=0, init=False)

    @staticmethod
    def convert(value: str, format: str) -> int:
        return int(value, 10)


@dataclass
class BoolFieldSpec:
    name: str
    help: str
    type: str = field(default="boolean", init=False)
    zero: int = field(default=0, init=False)

    @staticmethod
    def convert(value: str, format: str) -> bool:
        return value.strip().lower() in ("true", "yes", "1")


@dataclass
class DecimalFieldSpec:
    name: str
    help: str
    type: str = field(default="number", init=False)
    zero: Decimal = field(default=Decimal("0"), init=False)

    @staticmethod
    def convert(value: str, format: str) -> Decimal:
        fmt = NumberFormat.from_format(format)
        return Decimal(fmt.normalize(value))


@dataclass
class DateFieldSpec:
    name: str
    help: str
    type: str = field(default="date", init=False)
    zero: None = field(default=None, init=False)

    @staticmethod
    def convert(value: str, format: str) -> datetime.date:
        # see https://docs.python.org/3/library/time.html#time.strptime
        if format != "":
            dt = parse_datetime(value, common_date_formats.get(format, format))
        elif is_number(value):
            ts = float(value)
            # will not allow for dates before 1970-05-08, which is acceptable
            while ts > 1.1e10:
                ts /= 1000.0
            dt = datetime.datetime.fromtimestamp(ts)

        else:
            dt = datetime.datetime.fromisoformat(value)

        return dt.date()  # if dt is not None else None


@dataclass
class TimeFieldSpec:
    name: str
    help: str
    type: str = field(default="time", init=False)
    zero: None = field(default=None, init=False)

    @staticmethod
    def convert(value: str, format: str) -> datetime.time:
        # see https://docs.python.org/3/library/time.html#time.strptime
        dt = datetime.datetime.strptime(value, common_time_formats.get(format, format))
        assert dt.date() == datetime.date(1900, 1, 1)

        return dt.time()


@dataclass
class DateTimeFieldSpec:
    name: str
    help: str
    type: str = field(default="datetime", init=False)
    zero: None = field(default=None, init=False)

    @staticmethod
    def convert(value: str, format: str) -> datetime.datetime:
        # see https://docs.python.org/3/library/time.html#time.strptime
        if format != "":
            dt = parse_datetime(value, common_datetime_formats.get(format, format))
        elif is_number(value):
            ts = float(value)
            # will not allow for dates before 1970-05-08, which is acceptable
            while ts > 1.1e10:
                ts /= 1000.0
            dt = datetime.datetime.fromtimestamp(ts)
        else:
            dt = datetime.datetime.fromisoformat(value)

        return dt


class FieldSpecRegistry:
    """Registry for FieldSpec classes that"""

    def __init__(self):
        self.known_specs: dict[str, type[dataclass]] = {}

    def add(self, spec: type[dataclass]) -> None:
        assert_field_spec_class(spec)
        self.known_specs[spec.type] = spec

    def get_types(self) -> list[str]:
        return list(sorted(self.known_specs.keys()))

    def make_spec_list(self, data: list[dict[str, str]]) -> list[FieldSpec]:
        result: list[FieldSpec] = []
        for idx, item in enumerate(data, 1):
            assert "name" in item, f"Item {idx} requires a field 'name'."
            assert "type" in item, f"Item {idx} reqrequires a field 'type'."
            types = self.get_types()
            name, type = item["name"], item["type"]
            assert (
                item["type"] in types
            ), f"Item {idx} ('{name}') requests unknown type '{type}' - known types are {types}"
            kwargs = item.copy()
            del kwargs["type"]

            try:
                result.append(self.known_specs[type](**kwargs))
            except TypeError as e:
                raise AssertionError(f"Item {idx} ('{name}') cannot be initialized: {e}")

        return result


def add_builtin_fieldspecs_to_registry(registry: FieldSpecRegistry):
    for name, cls in globals().items():
        try:
            if not is_field_spec_class(cls):
                continue
            if not is_field_spec_class(cls):
                continue
            logger.info(f"Adding builtin {name} to fieldspec-registry.")
            registry.add(cls)

        except TypeError:
            pass



import datetime
from typing import Optional

from contablo.format_helpers import get_date_strptime_from_format
from contablo.format_helpers import get_time_strptime_from_format
from contablo.numberformat import NumberFormat


def dicts_equal_without_keys(a, b, ignore_keys):
    ka = set(a).difference(ignore_keys)
    kb = set(b).difference(ignore_keys)
    return ka == kb and all(a[k] == b[k] for k in ka)


def dicts_equal_in_keys(a, b, compared_keys):
    ka = set(a).intersection(compared_keys)
    kb = set(b).intersection(compared_keys)
    return ka == kb and all(a[k] == b[k] for k in ka)


def match_to_template(text: str, template: str, strict_whitespace: bool = False) -> Optional[dict[str, str]]:
    import re

    pattern = re.escape(template)
    if not strict_whitespace:
        pattern = re.sub(r"(\\\s)+", r"\\s+", pattern)
    pattern = re.sub(r"\\\{\\\}", r".+", pattern)
    pattern = re.sub(r"\\\{(\w+)\\\}", r"(?P<\1>.*)", pattern)
    # print(template, pattern)
    match = re.match(pattern, text)
    return match.groupdict() if match is not None else None


def split_after_prefix(prefix: str, text: str) -> list[str]:
    if text == prefix:
        return [prefix]
    if not text.startswith(prefix):
        return []
    delimiter = text[len(prefix)]
    return text.split(delimiter)


def check_condition_value_is_empty(raw: str, parts: list[str]) -> bool:
    assert len(parts) - 1 == 0, "[]"
    assert parts[0] == "empty"
    return not raw


def check_condition_value_is_not_empty(raw: str, parts: list[str]) -> bool:
    assert len(parts) - 1 == 0, "[]"
    assert parts[0] == "notempty"
    return (raw is not None) and (raw != "")


def check_condition_value_compare(raw: str, parts: list[str]) -> bool:
    assert len(parts) - 1 == 3, "[value, type, format] with type one of [number, date, time, datetime]"
    mode, raw_ref, subtype, format = parts
    assert subtype in ["number", "date", "time", "datetime"], f"unknwon type <{subtype}>"

    if raw is None:
        return False

    if subtype == "number":
        num_fmt = NumberFormat.from_format(format)
        value = float(num_fmt.normalize(raw))
        ref_value = float(num_fmt.normalize(raw_ref))
    elif subtype == "date":
        date_fmt = get_date_strptime_from_format(format)
        value = datetime.datetime.strptime(raw, date_fmt).date()
        ref_value = datetime.datetime.strptime(raw_ref, date_fmt).date()
    elif subtype == "time":
        time_fmt = get_time_strptime_from_format(format)
        value = datetime.datetime.strptime(raw, time_fmt).time()
        ref_value = datetime.datetime.strptime(raw_ref, time_fmt).time()
    else:
        raise NotImplementedError(f"check_condition_value_compare requires implementation for {subtype=}")
    cls = value.__class__
    methods = {"=": cls.__eq__, "!=": cls.__ne__, "<": cls.__lt__, ">": cls.__gt__, "<=": cls.__le__, ">=": cls.__ge__}
    assert mode in methods, f"Unknown condition mode '{mode}'"
    func = methods[mode]
    # print(f"## check if {value} {mode} {ref_value} using {func}")
    return func(value, ref_value)


def check_condition_string_is(raw: str, parts: list[str]) -> bool:
    assert 1 <= len(parts) - 1 <= 2, "[text, [i]]"

    _, text, flags = (parts + [""])[:3]

    assert all([c in "i" for c in flags]), f"Unsupported flag in <{flags}>"

    if "i" in flags:
        raw, text = raw.lower(), text.lower()

    return raw == text


def check_condition_string_is_not(raw: str, parts: list[str]) -> bool:
    return not check_condition_string_is(raw, parts)


def check_condition(cond: str, raw: str) -> list[str]:
    # Todo: maybe put this in a class, allow to query for conditions with syntax (see assert in handlers)
    conditions = [
        ["empty", 0, check_condition_value_is_empty],
        ["notempty", 0, check_condition_value_is_not_empty],
        ["=", 3, check_condition_value_compare],
        ["!=", 3, check_condition_value_compare],
        ["<", 3, check_condition_value_compare],
        [">", 3, check_condition_value_compare],
        ["<=", 3, check_condition_value_compare],
        [">=", 3, check_condition_value_compare],
        ["is", 1, check_condition_string_is],  # value
        ["is", 2, check_condition_string_is],  # value, [i]
        ["not", 1, check_condition_string_is_not],  # value
        ["not", 2, check_condition_string_is_not],  # value, [i]
    ]
    possible_cause = ""
    for type, nargs, func in conditions:
        if not (parts := split_after_prefix(type, cond)):
            continue
        if len(parts) - 1 != nargs:
            possible_cause = f"Expected {nargs} argument{'s' if nargs == 1 else ''} for '{type}', got {parts[1:]}."
            continue
        return func(raw, parts)
    if possible_cause:
        raise ValueError(possible_cause)
    raise ValueError("Unknown condition {type}.")


def check_conditions(
    conditions: dict[str, str],
    primary_raw_values: dict[str, str],
    secondary_raw_values: dict[str, str] | None = None,
) -> bool:
    """Check if the provided raw values meet all conditions. Raises KeyError for unknown condition keys."""
    secondary_raw_values = secondary_raw_values or {}
    for field, cond in conditions.items():
        # might raise a KeyError indicating an unknown field or input key
        if field not in primary_raw_values and field not in secondary_raw_values:
            keys = set(list(primary_raw_values.keys()) + list(secondary_raw_values.keys()))
            raise KeyError(f"Field <{field}> requested when only <{keys}> where given.")
        if not check_condition(cond, primary_raw_values.get(field, secondary_raw_values.get(field, None))):
            return False
    return True
